#include <vector>
#include <string>

#include "wx/wxprec.h"
#include "wx/colordlg.h"
#include "wx/filedlg.h"
#include "wx/splitter.h"
#include "wx/treectrl.h"
#include "wx/grid.h"
#include "wx/tokenzr.h"
#include "wx/fontdlg.h"
#include "wx/image.h"
#include "wx/dnd.h"
#include "wx/notebook.h"
#include "wx/html/htmlwin.h"

#include "api.h"

enum { WINDOW_TYPE='WX', TOPWIN_TYPE, SIZER_TYPE, SPLITTER_TYPE, GRID_TYPE, TREE_TYPE, NOTEBOOK_TYPE, MENUBAR_TYPE, HTML_TYPE};

#pragma warning(disable : 4100)

// Set one event handle for an widget to call Z-Script to handle events

std::string wxs2str(const wxString &wxs)
{
	std::string str;
	size_t n = wcstombs(NULL, wxs, wxs.Len());
	str.resize(n+10);
	wcstombs((char*)str.c_str(), wxs, wxs.Len());
	return str;
}

class myEvtHandler : public wxEvtHandler
{
	void *_ctx, *_func;
	wxWindow *_obj;
public:
	myEvtHandler(void *ctx, void *func, wxWindow *obj) : _ctx(ctx), _func(func), _obj(obj) { }

	~myEvtHandler() { }

	void OnCommand(wxCommandEvent& event);

	void OnKey(wxKeyEvent& event);

	void OnMouse(wxMouseEvent& event);

	void OnSize(wxSizeEvent& event);

	void OnPaint(wxPaintEvent& event);

	void OnSelect(wxCommandEvent& event);

	void PopEvent() { _obj->PopEventHandler(); }
};

// main frame

class myFrame : public wxFrame
{
	void *_ctx, *_func;

	std::vector<myEvtHandler*> _evth;

	wxTimer _timer;

	wxWindow _window;

public:
	myFrame() : wxFrame(NULL, wxID_ANY, "http://www.zegraph.com/z-script/"), _ctx(0), _func(0)
	{
		_timer.SetOwner(this, wxID_ANY);
		_window.Hide();
		DragAcceptFiles(true);
	}

	virtual ~myFrame()
	{
		if (_ctx && _func) {
			void *p[1];
			p[0] =  api_create_string(_ctx, "exit");
			api_call_func(_ctx, _func, 1, p);
		}
		int i;
		for (i = 0; i < _evth.size(); i++) {
			_evth[i]->PopEvent();
		}
		for (i = 0; i < _evth.size(); i++) {
			delete _evth[i];
		}
	}

	void OnSize(wxSizeEvent& event);

	void OnTimer(wxTimerEvent& event);

	void OnPaint(wxPaintEvent& event);

	void OnDropFile(wxDropFilesEvent& event);

    void OnDrag(wxCommandEvent& event);

	void OnActivate(wxActivateEvent& event);
 
	void OnErase(wxEraseEvent& event);

	void OnFocus(wxFocusEvent& event);

	wxTimer &timer() { return _timer; }

	void evth(myEvtHandler *h) { _evth.push_back(h); }

	void callback(void *ctx, void *func) { _ctx = ctx; _func = func; }

	void *ctx() { return _ctx; }

	void *func() { return _func; }

	wxWindow *window() { return &_window; }

private:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(myFrame, wxFrame)
	EVT_SIZE(myFrame::OnSize)
	EVT_TIMER(wxID_ANY, myFrame::OnTimer)
	EVT_PAINT(myFrame::OnPaint)
	EVT_DROP_FILES(myFrame::OnDropFile)
	EVT_ACTIVATE(myFrame::OnActivate)
	EVT_ERASE_BACKGROUND(myFrame::OnErase)
END_EVENT_TABLE()


class myStaticText : public wxStaticText
{
	wxBitmap _bmp;
public:
	myStaticText(wxWindow* parent,  wxWindowID id,  const wxString& label,
		const wxPoint& pos=wxDefaultPosition,  const wxSize& size=wxDefaultSize, long style=0, const wxString& name=wxStaticTextNameStr) {
		Create(parent, id, label, pos, size, style, name);
		SetBackgroundColour(parent->GetBackgroundColour());
		SetForegroundColour(parent->GetForegroundColour());
		SetWindowStyle(wxTRANSPARENT_WINDOW|GetWindowStyle());
	} 

	virtual void OnPaint(wxPaintEvent& event);  

private:
    DECLARE_EVENT_TABLE()
};


BEGIN_EVENT_TABLE(myStaticText, wxStaticText)
	EVT_PAINT(myStaticText::OnPaint)
END_EVENT_TABLE()

void myStaticText::OnPaint(wxPaintEvent& evt) {
	wxPaintDC dc(this);
	dc.SetFont(GetFont());
	dc.SetTextForeground(GetForegroundColour());
	dc.SetTextBackground(GetBackgroundColour());
	dc.DrawText(GetLabel(), 0, 0);
}

// Application startup

class myApp : public wxApp
{
	void *_module;
public:
	myApp() : _module(0) { }

	~myApp() { if (_module) api_delete_module(_module); }

	virtual bool OnInit();

	virtual int OnExit() { return 0; }
};

IMPLEMENT_APP(myApp)


bool myApp::OnInit()
{
	myFrame *frame = new myFrame;
	frame->SetIcon(wxICON(sample));
	SetTopWindow(frame);
	// call Z-Script to create widgets

	std::string str = wxs2str(argv[0]);
	if (argc > 1) str = wxs2str(argv[1]);
	int i, k1 = 0, k2 = str.size();
	for (i = 0; i < str.size(); i++) {
		if (str.at(i) == '\\' || str.at(i) == '/') k1 = i + 1;
		if (str.at(i) == '.')  k2 = i;
	}
	str = str.substr(k1, k2-k1);
	char err[256];
	err[0] = 0;
	_module = api_parse_file(str.c_str(), err);
	if (!_module) {
		wxMessageBox(err, "Z-Script", wxOK | wxICON_ERROR);
		return false;
	}
	if (api_exec_module(_module, err) <= 0) {
		wxMessageBox(err, "Z-Script", wxOK | wxICON_ERROR);
		return false;
	}
	frame->Show(true);
	return true;
}

class myTreeCtrl : public wxTreeCtrl
{
	void *_ctx, *_func;

	wxImageList	_img;

public:
	myTreeCtrl(wxWindow* parent, void *ctx, void *func) : wxTreeCtrl(parent, wxID_ANY), _ctx(ctx), _func(func)
	{
		SetImageList(&_img);
	}

	virtual ~myTreeCtrl() { }
	
    void OnActivated(wxTreeEvent& event);

    void OnRightClick(wxTreeEvent& event);

    void OnBeginDrag(wxTreeEvent& event);

    void OnEndDrag(wxTreeEvent& event);

	int addImage(const char *fname);

private:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(myTreeCtrl, wxTreeCtrl)
	EVT_TREE_ITEM_ACTIVATED(wxID_ANY, myTreeCtrl::OnActivated) 
	EVT_TREE_ITEM_RIGHT_CLICK(wxID_ANY, myTreeCtrl::OnRightClick) 
	EVT_TREE_BEGIN_DRAG(wxID_ANY, myTreeCtrl::OnBeginDrag)
	EVT_TREE_END_DRAG(wxID_ANY, myTreeCtrl::OnEndDrag)
END_EVENT_TABLE()


int myTreeCtrl::addImage(const char *fname)
{
	wxImage img;
	img.AddHandler(new wxPNGHandler);
	img.AddHandler(new wxGIFHandler);
	img.AddHandler(new wxICOHandler);
	img.AddHandler(new wxJPEGHandler);
	img.AddHandler(new wxTIFFHandler);
	if (img.LoadFile(fname, wxBITMAP_TYPE_ANY)) {
		wxBitmap bmp(img);
		return _img.Add(bmp);
	}
	return -1;
}


//////// Z-Script ///////////////////////////////////////////////////

struct zsOptions
{
	const char *name;
	long code;
};

// utility functions

static long get_options(void *ctx, const char *str, struct zsOptions ops[])
{
	long flag = 0;
	int k1 = 0, k2, n = strlen(str);
	char buf[128];
	for (k2 = 1; k2 <= n; k2++) {
		if (str[k2] == '|' || k2 == n) {
			strncpy(buf, str+k1, k2-k1);
			buf[k2-k1] = 0;
			int i = 0;
			while (ops[i].name != 0) {
				if (strcmp(buf, ops[i].name) == 0) {
					flag |= ops[i].code;
					break;
				}
				i++;
			}
			if (ops[i].name == 0) api_runtime_error(ctx, "incorrect option string");
			k2++;
			k1 = k2;
		}
	}
	return flag;
}

// none-type functions

static void* wx_message(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxMessageBox(api_get_string(ctx, args[0]), "Z-Script", wxOK, NULL);
	return 0;
}

static void* wx_frame(void *ctx, int nargs, void** args)
{
	myFrame *frame = (myFrame*)wxGetApp().GetTopWindow();
	return api_create_user(ctx, frame, 0, 0, TOPWIN_TYPE);
}

static void* wx_cast(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	return api_create_user(ctx, api_get_ptr(ctx, args[0]), 0, 0, WINDOW_TYPE);
}

static void* wx_active(void *ctx, int nargs, void** args)
{
	return api_create_integer(ctx, wxGetApp().IsActive());
}

static void* wx_size(void *ctx, int nargs, void** args)
{
	int w, h;
	wxDisplaySize(&w, &h);
	void *arr = api_create_array(ctx, 2);
	api_set_array_object(ctx, arr, "0", api_create_integer(ctx, w));
	api_set_array_object(ctx, arr, "1", api_create_integer(ctx, h));
	return arr;
}


static void* wx_split(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxString path, name, ext;
	wxFileName::SplitPath(api_get_string(ctx, args[0]), NULL, &path, &name, &ext);
	void *arr = api_create_array(ctx, 3);
	std::string str = wxs2str(path);
	api_set_array_object(ctx, arr, "0", api_create_string(0,str.c_str()));
	str = wxs2str(name);
	api_set_array_object(ctx, arr, "1", api_create_string(0,str.c_str()));
	str = wxs2str(ext);
	api_set_array_object(ctx, arr, "2",  api_create_string(0,str.c_str()));
	return arr;
}

static void* wx_exec(void* ctx, int nargs, void **args)
{
	if (nargs < 1) api_input_error(ctx);
	const char *cmd = api_get_string(ctx, args[0]);
	wxExecute(cmd, wxEXEC_ASYNC);
	return 0;
}

static void* wx_exit(void *ctx, int nargs, void** args)
{
	myApp& app = wxGetApp();
	app.ExitMainLoop();
	return 0;
}

// frame/dialog functions

static void* topwin_timer(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	int ms = api_get_integer(ctx, args[1]);
	if (ms <= 0)
		frame->timer().Stop();
	else
		frame->timer().Start(ms);
	return 0;
}

static void* topwin_dirdialog(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	const char *path = "";
	if (nargs > 1) path = api_get_string(ctx, args[1]);
	wxDirDialog dialog(frame, "Z-Script", path);
	if (dialog.ShowModal() == wxID_OK) {
		std::string str = wxs2str(dialog.GetPath());
		return api_create_string(ctx, str.c_str());
	}
	return 0;
}

static void* topwin_filedialog(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	const char *path = "";
	if (nargs > 1) path = api_get_string(ctx, args[1]);
	int flag = 0;
	if (nargs > 2) flag = api_get_integer(ctx, args[2]);
	if (flag != 0) {
		wxFileDialog dialog(frame, "Z-Script",  path, "", "*.*", wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
		if (dialog.ShowModal() == wxID_OK) {
			std::string str = wxs2str(dialog.GetPath());
			return api_create_string(ctx, str.c_str());
		}
	}
	else {
		wxFileDialog dialog(frame, "Z-Script",  path, "", "*.*", wxFD_OPEN | wxFD_MULTIPLE);
		if (dialog.ShowModal() == wxID_OK) {
			wxArrayString ss;
			dialog.GetPaths(ss);
			if (ss.Count() == 1) {
				std::string str = wxs2str(ss.Item(0));
				return api_create_string(ctx, str.c_str());
			}
			void *arr = api_create_array(ctx, ss.Count());
			char key[64];
			for (unsigned i = 0; i < ss.Count(); i++) {
				sprintf(key, "%d", i);
				std::string str = wxs2str(ss.Item(i));
				api_set_array_object(ctx, arr, key, api_create_string(0, str.c_str()));
			}
			return arr;
		}
	}
	return 0;
}

static void* topwin_colordialog(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	wxColourDialog dialog(frame, NULL);
	if (dialog.ShowModal() == wxID_OK) {
		wxColour color = dialog.GetColourData().GetColour();
		void *arr = api_create_array(ctx, 3);
		api_set_array_object(ctx, arr, "0", api_create_integer(0, color.Red()));
		api_set_array_object(ctx, arr, "1", api_create_integer(0, color.Green()));
		api_set_array_object(ctx, arr, "2", api_create_integer(0, color.Blue()));
		return arr;
	}
	return 0;
}

static void* topwin_fontdialog(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	wxFontDialog dialog(frame);
	if (dialog.ShowModal() == wxID_OK) {
		wxFont font = dialog.GetFontData().GetChosenFont(); 
		void *arr = api_create_array(ctx, 7);
		std::string str = wxs2str(font.GetFaceName());
		api_set_array_object(ctx, arr, "0", api_create_integer(0, font.GetDefaultEncoding()));
		api_set_array_object(ctx, arr, "1", api_create_integer(0, font.GetFamily()));
		api_set_array_object(ctx, arr, "2", api_create_string(0, str.c_str()));
		api_set_array_object(ctx, arr, "3", api_create_integer(0, font.GetPointSize()));
		api_set_array_object(ctx, arr, "4", api_create_integer(0, font.GetStyle()));
		api_set_array_object(ctx, arr, "5", api_create_integer(0, font.GetUnderlined()));
		api_set_array_object(ctx, arr, "6", api_create_integer(0, font.GetWeight()));
		return arr;
	}
	return 0;
}

static void* topwin_title(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	frame->SetTitle(api_get_string(ctx, args[1]));
	return 0;
}

void* topwin_icon(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	wxString fname(api_get_string(ctx, args[1]));
	wxIcon ico;
	if (!ico.LoadFile(fname, wxBITMAP_TYPE_ICO)) {
		if (!ico.LoadFile(fname, wxBITMAP_TYPE_GIF)) {
			if (!ico.LoadFile(fname, wxBITMAP_TYPE_PNG)) {
				if (!ico.LoadFile(fname, wxBITMAP_TYPE_XPM)) {
					if (!ico.LoadFile(fname, wxBITMAP_TYPE_XBM))
						api_runtime_error(ctx, "unsupported icon type");
				}
			}
		}
	}
	frame->SetIcon(ico);
	return 0;
}

static void* topwin_size(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *w = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	int flag = api_get_integer(ctx, args[1]);
	if (flag < 0)
		w->ShowFullScreen(true);
	else if (flag == 0)
		w->Maximize(false);
	else
		w->Maximize(true);
	return 0;
}

static void* topwin_menubar(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	wxMenuBar *menuBar = frame->GetMenuBar();
	if (!menuBar) {
		menuBar = new wxMenuBar();
		frame->SetMenuBar(menuBar);
	}
	if (nargs == 1) {
		return api_create_user(ctx,menuBar,0,0,MENUBAR_TYPE);
	}
	wxMenu *menu = new wxMenu;
	menuBar->Append(menu, api_get_string(ctx, args[1]));
	void *ret = api_create_array(ctx, nargs-2);
	wxMenu *subm = 0;
	bool check = false;
	bool radio = false;
	for (int i = 2, j = 0; i < nargs; i++) {
		const char* s = api_get_string(ctx, args[i]);
		if (s[0] == '^') {
			check = !check;
			radio = false;
			continue;
		}
		if (s[0] == '@') {
			radio = !radio;
			check = false;
			continue;
		}
		if (s[0] == '>') {
			if (i == nargs-1 || subm) api_input_error(ctx);
			subm = new wxMenu;
			menu->Append(wxID_ANY, api_get_string(ctx, args[i+1]), subm);
			i++;
			continue;
		}
		if (s[0] == '<') {
			if (!subm) api_input_error(ctx);
			subm = 0;
			continue;
		}
		if (s[0] == '_') {
			check = false;
			radio = false;
			if (subm)
				subm->AppendSeparator();
			else
				menu->AppendSeparator();
			continue;
		}
		long id = wxNewId();
		api_set_array_object2(ctx, ret, j++, api_create_integer(0, id));
		wxRegisterId(id);
		if (subm) {
			if (check) {
				subm->AppendCheckItem(id, s);
			}
			else if (radio) {
				subm->AppendRadioItem(id, s);
			}
			else {
				subm->Append(id, s);
			}
		}
		else {
			if (check) {
				menu->AppendCheckItem(id, s);
			}
			else if (radio) {
				menu->AppendRadioItem(id, s);
			}
			else {
				menu->Append(id, s);
			}
		}
	}
	return ret;
}

static void* topwin_toolbar(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
    frame->SetToolBar(NULL);
	wxToolBar *toolBar = frame->CreateToolBar();
	void *arr = api_create_array(ctx, 1+nargs/2);
	for (int i = 2, j = 0; i < nargs; i += 2) {
		const char *fname = api_get_string(ctx, args[i-1]);
		const char* tip = api_get_string(ctx, args[i]);
		if (strcmp(fname, "|") == 0) {
			toolBar->AddSeparator();
			continue;
		}
		wxImage img;
		img.AddHandler(new wxPNGHandler);
		img.AddHandler(new wxGIFHandler);
		img.AddHandler(new wxICOHandler);
		if (!img.LoadFile(wxString(fname))) api_runtime_error(ctx, "unsupported image type");
		wxBitmap bmp(img);
		long id = wxNewId();
		api_set_array_object2(ctx, arr, j++, api_create_integer(0, id));
		wxRegisterId(id);
		toolBar->AddTool(id, wxEmptyString, bmp, tip);
	}
//    toolBar->Realize();
	return arr;
}

static void* topwin_statusbar(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	int n = api_get_integer(ctx, args[1]);
	if (nargs == 2) {
		int width[10] = { 0, -1, -1, -1, -1, -1, -1, -1, -1, -1 };
		if (n < 1) n = 1;
		if (n > 10) n = 10;
		frame->SetStatusBar(NULL);
		frame->CreateStatusBar(n+1);
		frame->SetStatusWidths(n, width);
	}
	else {
		frame->SetStatusText(api_get_string(ctx, args[2]), n+1);
	}
	return 0;
}

static void* topwin_push(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[1], WINDOW_TYPE);
	w->Reparent(frame->window());
	w->Hide();
	return 0;
}

static void* topwin_pop(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[1], WINDOW_TYPE);
	w->Reparent(frame);
	w->Show();
	return 0;
}

static void* topwin_update(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	frame->Update();
	return 0;
}

static void* topwin_callback(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myFrame *frame = (myFrame*)api_get_user(ctx, args[0], TOPWIN_TYPE);
	frame->callback(ctx, api_get_func(ctx, api_get_string(ctx, args[1])));
	return 0;
}

// window function

static void* win_enable(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	w->Enable(true);
	return 0;
}

static void* win_disable(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	w->Enable(false);
	return 0;
}

static void* win_id(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	return api_create_integer(ctx, w->GetId());
}

static void* win_close(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	w->Close(true);
	return 0;
}

static void* win_hwnd(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	return api_create_user(ctx, w->GetHandle(), 0, 0, 0);
}

static void* win_size(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (nargs == 1) {
		void *arr = api_create_array(ctx, 3);
		int w, h;
		win->GetClientSize(&w, &h);
		api_set_array_object(ctx, arr, "0", api_create_integer(0, w));
		api_set_array_object(ctx, arr, "1", api_create_integer(0, h));
		return arr;
	}
	else {
		if (nargs < 3) api_input_error(ctx);
		if (win->IsKindOf(CLASSINFO(wxFrame))) {
			win->SetClientSize(api_get_integer(ctx, args[1]), api_get_integer(ctx, args[2]));
		}
		else {
			win->SetSize(api_get_integer(ctx, args[1]), api_get_integer(ctx, args[2]));
		}
	}
	return 0;
}

static void* win_get(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (win->IsKindOf(CLASSINFO(wxSlider))) {
		return api_create_integer(ctx, ((wxSlider*)win)->GetValue());
	}
	if (win->IsKindOf(CLASSINFO(wxCheckBox))) {
		return api_create_integer(ctx, ((wxCheckBox*)win)->GetValue());
	}
	if (win->IsKindOf(CLASSINFO(wxRadioBox))) {
		std::string str = wxs2str(((wxRadioBox*)win)->GetStringSelection());
		return api_create_string(ctx, str.c_str());
	}
	if (win->IsKindOf(CLASSINFO(wxComboBox))) {
		std::string str = wxs2str(((wxComboBox*)win)->GetValue());
		return api_create_string(ctx, str.c_str());
	}
	std::string str = wxs2str(win->GetLabel());
	return api_create_string(ctx, str.c_str());
}

static void* win_set(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (win->IsKindOf(CLASSINFO(wxSlider))) {
		((wxSlider*)win)->SetValue(api_get_integer(ctx, args[1]));
		return 0;
	}
	if (win->IsKindOf(CLASSINFO(wxRadioButton))) {
		((wxRadioButton*)win)->SetValue(api_get_integer(ctx, args[1]) != 0);
		return 0;
	}
	if (win->IsKindOf(CLASSINFO(wxCheckBox))) {
		((wxCheckBox*)win)->SetValue(api_get_integer(ctx, args[1]) != 0);
		return 0;
	}
	if (win->IsKindOf(CLASSINFO(wxComboBox))) {
		wxComboBox *p = (wxComboBox*)win;
		int n = api_get_array_size(ctx, args[1]);
		p->Remove(0,p->GetCount());
		for (int i = 0; i < n; i++) {
			const char *s = api_get_string(ctx, api_get_array_object2(ctx, args[1], i));
			p->Append(wxString(s));
		}
		int sel = 0;
		if (nargs > 2) sel = api_get_integer(ctx, args[2]);
		p->Select(sel);
		return 0;
	}
	win->SetLabel(api_get_string(ctx, args[1]));
	return 0;
}

static void* win_bgcolor(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (nargs == 1) {
		wxColour color = w->GetBackgroundColour();
		void *arr = api_create_array(ctx, 3);
		api_set_array_object(ctx, arr, "0", api_create_integer(0, color.Red()));
		api_set_array_object(ctx, arr, "1", api_create_integer(0, color.Green()));
		api_set_array_object(ctx, arr, "2", api_create_integer(0, color.Blue()));
		return arr;
	}
	if (nargs < 4) api_input_error(ctx);
	int r = api_get_integer(ctx, args[1]);
	int g = api_get_integer(ctx, args[2]);
	int b = api_get_integer(ctx, args[3]);
	wxColour color(r, g, b);
	w->SetBackgroundColour(color);
	return 0;
}

static void* win_fgcolor(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (nargs == 1) {
		wxColour color = w->GetForegroundColour();
		void *arr = api_create_array(ctx, 3);
		api_set_array_object(ctx, arr, "0", api_create_integer(0, color.Red()));
		api_set_array_object(ctx, arr, "1", api_create_integer(0, color.Green()));
		api_set_array_object(ctx, arr, "2", api_create_integer(0, color.Blue()));
		return arr;
	}
	if (nargs < 4) api_input_error(ctx);
	int r = api_get_integer(ctx, args[1]);
	int g = api_get_integer(ctx, args[2]);
	int b = api_get_integer(ctx, args[3]);
	wxColour color(r, g, b);
	w->SetForegroundColour(color);
	return 0;
}

static void* win_focus(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	w->SetFocus();
	return 0;
}

static void* win_font(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (nargs == 1) {
		wxFontDialog dialog(w);
		if (dialog.ShowModal() == wxID_OK) {
			wxFont font = dialog.GetFontData().GetChosenFont(); 
			w->SetFont(font);
		}
		return 0;
	}
	wxFont font = w->GetFont();
	font.SetPointSize(api_get_integer(ctx, args[1]));
	w->SetFont(font);
	return 0;
}

static void* win_panel(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxPanel *p = new wxPanel(w);
	if (nargs > 1) {
		p->SetAutoLayout(api_get_integer(ctx,args[1])!=0);
	}
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_sizer(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *type = api_get_string(ctx, args[1]);
	if (strcmp(type, "horbox") == 0) {
		wxBoxSizer *sizer = new wxBoxSizer(wxHORIZONTAL);
		w->SetSizer(sizer);
		sizer->SetSizeHints(w);   
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	if (strcmp(type, "verbox") == 0) {
		wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
		w->SetSizer(sizer);
		sizer->SetSizeHints(w);   
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	if (strcmp(type, "horstatic") == 0) {
		if (nargs < 5) api_input_error(ctx);
		wxStaticBoxSizer *sizer = new wxStaticBoxSizer(wxHORIZONTAL, w, api_get_string(ctx, args[2]));
		sizer->SetMinSize(api_get_integer(ctx, args[3]), api_get_integer(ctx, args[4]));
		w->SetSizer(sizer);
		sizer->SetSizeHints(w);   
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	if (strcmp(type, "verstatic") == 0) {
		if (nargs < 5) api_input_error(ctx);
		wxStaticBoxSizer *sizer = new wxStaticBoxSizer(wxVERTICAL, w, api_get_string(ctx, args[2]));
		sizer->SetMinSize(api_get_integer(ctx, args[3]), api_get_integer(ctx, args[4]));
		w->SetSizer(sizer);
		sizer->SetSizeHints(w);   
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	if (strcmp(type, "flexgrid") == 0) {
		if (nargs < 6) api_input_error(ctx);
		int rows = api_get_integer(ctx, args[2]);
		int vgap = api_get_integer(ctx, args[3]);
		int cols = api_get_integer(ctx, args[4]);
		int hgap = api_get_integer(ctx, args[5]);
		wxFlexGridSizer *sizer = new wxFlexGridSizer(rows, cols, vgap, hgap);
		w->SetSizer(sizer);
		sizer->SetSizeHints(w);   
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	api_runtime_error(ctx, "unsupported sizer type");
	return 0;
}

static void* win_style(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	zsOptions ops[] = { { "simple", wxSIMPLE_BORDER },
						{ "double", wxDOUBLE_BORDER },
						{ "sunken", wxSUNKEN_BORDER },
						{ "raised", wxRAISED_BORDER },
						{ "static", wxSTATIC_BORDER },
						{ "none", wxNO_BORDER },
						{ "vscroll", wxVSCROLL },
						{ "hscroll", wxHSCROLL },
						{ "clip", wxCLIP_CHILDREN },
						{ "transparent", wxTRANSPARENT_WINDOW },
						{ "fixed", (!wxCAPTION & !wxMINIMIZE_BOX & !wxMAXIMIZE_BOX & !wxRESIZE_BORDER) },
						{ 0, 0 } };
	long style = get_options(ctx, api_get_string(ctx, args[1]), ops);
	if (style == wxTRANSPARENT_WINDOW) {
		wxWindow *parent = w->GetParent();
		w->SetBackgroundColour(parent->GetBackgroundColour());
		w->SetForegroundColour(parent->GetForegroundColour());
	}
	w->SetWindowStyle(style);
	return 0;
}

static void* win_button(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *s = api_get_string(ctx, args[1]);
	wxSize size = wxDefaultSize;
	if (nargs > 3) {
		size.SetWidth(api_get_integer(ctx, args[2]));
		size.SetHeight(api_get_integer(ctx, args[3]));
	}
	wxButton *p = new wxButton(w, wxID_ANY, s, wxDefaultPosition, size);
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_checkbox(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *s = api_get_string(ctx, args[1]);
	int flag = api_get_integer(ctx, args[2]);
	wxCheckBox *p = new wxCheckBox(w, wxID_ANY, s);
	if (flag != 0)
		p->SetValue(true);
	else
		p->SetValue(false);
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_combobox(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxComboBox *p = new wxComboBox(parent, wxID_ANY);
	int n = api_get_array_size(ctx, args[1]);
	for (int i = 0; i < n; i++) {
		const char *s = api_get_string(ctx, api_get_array_object2(ctx, args[1], i));
		p->Append(wxString(s));
	}
	int sel = 0;
	if (nargs > 2) sel = api_get_integer(ctx, args[2]);
	p->Select(sel);
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_radiobox(void *ctx, int nargs, void** args)
{
	if (nargs < 5) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *title = api_get_string(ctx, args[1]);
	long style = wxRA_SPECIFY_COLS;
	const char *flag = api_get_string(ctx, args[2]);
	if (flag[0] == 'r' || flag[0] == 'R') style = wxRA_SPECIFY_ROWS;
	wxArrayString choices;
	for (int i = 3; i < nargs; i++) choices.Add(api_get_string(ctx, args[i]));
	wxRadioBox *p = new wxRadioBox(parent, wxID_ANY, title, wxDefaultPosition, wxDefaultSize, choices, 0, style);
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_html(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxHtmlWindow *p = new wxHtmlWindow(parent);
	return api_create_user(ctx, p, 0, 0, HTML_TYPE);
}

static void* win_slider(void *ctx, int nargs, void** args)
{
	if (nargs < 5) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *s = api_get_string(ctx, args[1]);
	long style = wxSL_LABELS|wxSL_AUTOTICKS;
	if (strcmp(s, "left") == 0) style |= wxSL_LEFT|wxSL_INVERSE;
	else if (strcmp(s, "right") == 0) style |= wxSL_RIGHT|wxSL_INVERSE;
	else if (strcmp(s, "top") == 0) style |= wxSL_TOP;
	else style |= wxSL_BOTTOM;
	int vmin  = api_get_integer(ctx, args[2]);
	int vmax  = api_get_integer(ctx, args[3]);
	int value = api_get_integer(ctx, args[4]);
	int freq = (vmax-vmin) / 10;
	if (nargs > 5) freq = api_get_integer(ctx, args[5]);
	wxSlider *p = new wxSlider(parent, wxID_ANY, value, vmin, vmax, wxDefaultPosition, wxDefaultSize, style);
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_statictext(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *s = api_get_string(ctx, args[1]);
	long style = wxALIGN_LEFT;
	if (nargs > 2) {
		int flag = api_get_integer(ctx, args[2]);
		if (flag > 0) style = wxALIGN_RIGHT;
		if (flag == 0) style = wxALIGN_CENTRE;
	}
	wxStaticText *p;
	if (nargs > 3 && api_get_integer(ctx, args[3]) != 0) {
		p = new myStaticText(parent, wxID_ANY, s, wxDefaultPosition, wxDefaultSize, style);
	}
	else {
		p = new wxStaticText(parent, wxID_ANY, s, wxDefaultPosition, wxDefaultSize, style);
	}
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_textbox(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxString value(api_get_string(ctx, args[1]));
	long style = 0;
	wxSize size = wxDefaultSize;
	if (nargs > 2) {
		if (api_is_string(args[2])) {
			zsOptions ops[] = { { "multiline", wxTE_MULTILINE },
								{ "password", wxTE_PASSWORD },
								{ "readonly", wxTE_READONLY },
								{ "left", wxTE_LEFT },
								{ "center", wxTE_CENTRE },
								{ "right", wxTE_RIGHT },
								{ 0, 0 } };
			style = get_options(ctx, api_get_string(ctx, args[2]), ops);
			if (nargs > 3) {
				size.SetWidth(api_get_integer(ctx, args[3]));
				if (nargs > 4) size.SetHeight(api_get_integer(ctx, args[4]));
			}
		}
		else {
			size.SetWidth(api_get_integer(ctx, args[2]));
			if (nargs > 3) size.SetHeight(api_get_integer(ctx, args[3]));
		}
	}
	wxTextCtrl *p = new wxTextCtrl(parent, wxID_ANY, value, wxDefaultPosition, size, style);
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* win_splitter(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxSplitterWindow *p = new wxSplitterWindow(parent, wxID_ANY);
	return api_create_user(ctx, p, 0, 0, SPLITTER_TYPE);
}

static void* win_grid(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxGrid *p = new wxGrid(parent, wxID_ANY);
	p->CreateGrid(api_get_integer(ctx, args[1]), api_get_integer(ctx, args[2]));
	return api_create_user(ctx, p, 0, 0, GRID_TYPE);
}

static void* win_treeview(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	void *func = 0;
	if (nargs > 1) func = api_get_func(ctx, api_get_string(ctx, args[1]));
	return api_create_user(ctx, new myTreeCtrl(parent, ctx, func), 0, 0, TREE_TYPE);
}

static void* win_notebook(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *parent = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	zsOptions ops[] = { { "top", wxNB_TOP },
						{ "left", wxNB_LEFT },
						{ "rigth", wxNB_RIGHT },
						{ "bottom", wxNB_BOTTOM },
						{ 0, 0 } };
	long style = wxNB_TOP;
	if (nargs > 1) style = get_options(ctx, api_get_string(ctx, args[1]), ops);
	wxNotebook *p = new wxNotebook(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, style);
	return api_create_user(ctx, p, 0, 0, NOTEBOOK_TYPE);
}

static void* win_bitmap(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	wxImage img;
	img.AddHandler(new wxPNGHandler);
	img.AddHandler(new wxGIFHandler);
	img.AddHandler(new wxICOHandler);
	img.AddHandler(new wxJPEGHandler);
	img.AddHandler(new wxTIFFHandler);
	if (!img.LoadFile(api_get_string(ctx, args[1]), wxBITMAP_TYPE_ANY)) {
		return api_create_integer(ctx, 0);
	}
	wxClientDC dc(win);
	wxCoord xdest=0, ydest=0, width, height;
	if (nargs > 4) {
		xdest = api_get_integer(ctx, args[2]);
		ydest = api_get_integer(ctx, args[3]);
		width = img.GetWidth();
		height = img.GetHeight();
	}
	else {
		dc.GetSize(&width, &height);
		img.Rescale(width, height);
	}
	wxBitmap bmp(img);
	wxMemoryDC mdc;
	mdc.SelectObject(bmp);
	dc.Blit(xdest, ydest, width, height, &mdc, 0, 0);
	return api_create_integer(ctx, 1);
}

static void* win_cursor(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	const char *flag = api_get_string(ctx, args[1]);
	if (strcmp(flag, "arrow") == 0) {
		wxCursor cursor(wxCURSOR_ARROW);
		win->SetCursor(cursor);
	}
	else if (strcmp(flag, "qarrow") == 0) {
		wxCursor cursor(wxCURSOR_QUESTION_ARROW);
		win->SetCursor(cursor);
	}
	else if (strcmp(flag, "hand") == 0) {
		wxCursor cursor(wxCURSOR_HAND);
		win->SetCursor(cursor);
	}
	else if (strcmp(flag, "ibeam") == 0) {
		wxCursor cursor(wxCURSOR_IBEAM);
		win->SetCursor(cursor);
	}
	else if (strcmp(flag, "noentry") == 0) {
		wxCursor cursor(wxCURSOR_NO_ENTRY);
		win->SetCursor(cursor);
	}
	else if (strcmp(flag, "wait") == 0) {
		wxCursor cursor(wxCURSOR_WAIT);
		win->SetCursor(cursor);
	}
	else {
		wxImage img(flag);
		wxCursor cursor(img);
		win->SetCursor(cursor);
	}
	return 0;
}

static void* win_refresh(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	win->Refresh(TRUE);
	win->Update();
	return 0;
}

static void* win_position(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (nargs == 1) {
		int x, y;
		win->GetPosition(&x, &y);
		void *arr = api_create_array(ctx, 2);
		api_set_array_object(ctx, arr, "0", api_create_integer(ctx, x));
		api_set_array_object(ctx, arr, "1", api_create_integer(ctx, y));
		return arr;
	}
	if (nargs < 3) api_input_error(ctx);
	wxPoint p(api_get_integer(ctx,args[1]),api_get_integer(ctx,args[2]));
	win->SetPosition(p);
	return 0;
}

static void* win_wrap(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxWindow *win = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	if (win->IsKindOf(CLASSINFO(wxStaticText))) {
		((wxStaticText*)win)->Wrap(api_get_integer(ctx,args[1]));
	}
	else if (win->IsKindOf(CLASSINFO(myStaticText))) {
		((myStaticText*)win)->Wrap(api_get_integer(ctx,args[1]));
	}
	return 0;
}

// HTML

static void* html_load(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxHtmlWindow *win = (wxHtmlWindow*)api_get_user(ctx, args[0], HTML_TYPE);
	win->LoadPage(api_get_string(ctx, args[1]));
	return 0;
}

static void* html_font(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxHtmlWindow *win = (wxHtmlWindow*)api_get_user(ctx, args[0], HTML_TYPE);
	double s0 = api_get_number(ctx, args[1]);
	int si[7];
	si[0] = int(s0);
	si[1] = int(1.2*s0);
	si[2] = int(1.4*s0);
	si[3] = int(1.6*s0);
	si[4] = int(1.8*s0);
	si[5] = int(2.0*s0);
	si[6] = int(2.2*s0);
	win->SetFonts("", "", si);
	return 0;
}

// menubar

static void* menubar_check(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxMenuBar *menuBar = (wxMenuBar*)api_get_user(ctx, args[0], MENUBAR_TYPE);
	if (nargs == 2) {
		int status=menuBar->IsChecked(api_get_integer(ctx, args[1]));
		// v3.01 bug?
		//return api_create_integer(ctx, status);
		return api_create_integer(ctx, status==0);
	}
	menuBar->Check(api_get_integer(ctx,args[1]), api_get_integer(ctx,args[2])!=0);
	return 0;
}

static void* menubar_enable(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxMenuBar *menuBar = (wxMenuBar*)api_get_user(ctx, args[0], MENUBAR_TYPE);
	if (nargs == 2) {
		int status=menuBar->IsEnabled(api_get_integer(ctx, args[1]));
		// v3.01 bug?
		// return api_create_integer(ctx, status);
		return api_create_integer(ctx, status==0);
	}
	menuBar->Enable(api_get_integer(ctx,args[1]), api_get_integer(ctx,args[2])!=0);
	return 0;
}

static void* menubar_label(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxMenuBar *menuBar = (wxMenuBar*)api_get_user(ctx, args[0], MENUBAR_TYPE);
	return api_create_string(ctx,menuBar->GetLabel(api_get_integer(ctx,args[1])));
}

// splitter

static void* splitter_min(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxSplitterWindow *s = (wxSplitterWindow*)api_get_user(ctx, args[0], SPLITTER_TYPE);
	int size = api_get_integer(ctx, args[1]);
	s->SetMinimumPaneSize(size);
	return 0;
}

static void* splitter_horizontal(void *ctx, int nargs, void** args)
{
	if (nargs < 4) api_input_error(ctx);
	wxSplitterWindow *s = (wxSplitterWindow*)api_get_user(ctx, args[0], SPLITTER_TYPE);
	wxWindow *w1 = (wxWindow*)api_get_user(ctx, args[1], WINDOW_TYPE);
	wxWindow *w2 = (wxWindow*)api_get_user(ctx, args[2], WINDOW_TYPE);
	int pos = api_get_integer(ctx, args[3]);
	s->SplitHorizontally(w1, w2, pos);
	return 0;
}

static void* splitter_vertical(void *ctx, int nargs, void** args)
{
	if (nargs < 4) api_input_error(ctx);
	wxSplitterWindow *s = (wxSplitterWindow*)api_get_user(ctx, args[0], SPLITTER_TYPE);
	wxWindow *w1 = (wxWindow*)api_get_user(ctx, args[1], WINDOW_TYPE);
	wxWindow *w2 = (wxWindow*)api_get_user(ctx, args[2], WINDOW_TYPE);
	int pos = api_get_integer(ctx, args[3]);
	w1->Reparent(s);
	w2->Reparent(s);
	s->SplitVertically(w1, w2, pos);
	return 0;
}

// grid control

static void* grid_get(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	std::string str = wxs2str(w->GetCellValue(api_get_integer(ctx, args[1]), api_get_integer(ctx, args[2])));
	return api_create_string(ctx, str.c_str());
}

static void* grid_set(void *ctx, int nargs, void** args)
{
	if (nargs < 4) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->SetCellValue(api_get_integer(ctx, args[1]), api_get_integer(ctx, args[2]), api_get_string(ctx, args[3]));
	return 0;
}

static void* grid_label(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->SetColLabelValue(api_get_integer(ctx, args[1]), api_get_string(ctx, args[2]));
	return 0;
}

static void* grid_rows(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	int n = w->GetNumberRows();
	if (nargs == 1) {
		return api_create_integer(ctx, n);
	}
	int m = api_get_integer(ctx, args[1]);
	if (m > n) {
		w->AppendRows(m - n);
	}
	return 0;
}

static void* grid_cols(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	int n = w->GetNumberCols();
	if (nargs == 1) {
		return api_create_integer(ctx, n);
	}
	int m = api_get_integer(ctx, args[1]);
	if (m > n) {
		w->AppendCols(m - n);
	}
	return 0;
}

static void* grid_clear(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->ClearGrid();
	return 0;
}

static void* grid_width(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->SetColSize(api_get_integer(ctx,args[1]),api_get_integer(ctx,args[2]));
	return 0;
}

static void* grid_height(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->SetRowSize(api_get_integer(ctx,args[1]),api_get_integer(ctx,args[2]));
	return 0;
}

static void* grid_enable(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->EnableEditing(true);
	return 0;
}

static void* grid_disable(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->EnableEditing(false);
	return 0;
}

static void* grid_readonly(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxGrid *w = (wxGrid*)api_get_user(ctx, args[0], GRID_TYPE);
	w->SetReadOnly(api_get_integer(ctx, args[1]), api_get_integer(ctx, args[2]));
	return 0;
}

// tree control

static void* tree_root(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	myTreeCtrl *w = (myTreeCtrl*)api_get_user(ctx, args[0], TREE_TYPE);
	int idx1 = -1;
	int idx2 = -1;
	if (nargs > 2) idx1 = w->addImage(api_get_string(ctx, args[2]));
	if (nargs > 3) idx2 = w->addImage(api_get_string(ctx, args[3]));
	w->AddRoot(api_get_string(ctx, args[1]), idx1, idx2);
	return 0;
}

static void* tree_append(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	myTreeCtrl *w = (myTreeCtrl*)api_get_user(ctx, args[0], TREE_TYPE);
	wxStringTokenizer tkz(api_get_string(ctx, args[1]), "/");
	wxTreeItemId id = w->GetRootItem();
	while ( tkz.HasMoreTokens()) {
		wxString token = tkz.GetNextToken();
		if (token.IsEmpty()) {
			id = w->GetRootItem();
		}
		else {
			bool flag = false;
			wxTreeItemIdValue cookie;
			wxTreeItemId id2 = w->GetFirstChild(id, cookie);
			while (id2.IsOk()) {
				wxString s = w->GetItemText(id2);
				if (s.Cmp(token) == 0) { flag = true; break; }
				id2 = w->GetNextChild(id, cookie);
			}
			if (!flag) {
				wxMessageBox("failed to find tree item", "Z-Script", wxOK | wxICON_ERROR);
				return 0;
			}
			id = id2;
		}
	}
	int idx1 = -1;
	int idx2 = -1;
	if (nargs > 3) idx1 = w->addImage(api_get_string(ctx, args[3]));
	if (nargs > 4) idx2 = w->addImage(api_get_string(ctx, args[4]));
	w->AppendItem(id, api_get_string(ctx, args[2]), idx1, idx2);
	return 0;
}

// notebook control

static void* notebook_add(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxNotebook *book = (wxNotebook*)api_get_user(ctx, args[0], NOTEBOOK_TYPE);
	wxPanel *p = new wxPanel(book);
	book->AddPage(p, api_get_string(ctx, args[1]), api_get_integer(ctx, args[2]));
	return api_create_user(ctx, p, 0, 0, WINDOW_TYPE);
}

static void* notebook_select(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxNotebook *book = (wxNotebook*)api_get_user(ctx, args[0], NOTEBOOK_TYPE);
	book->ChangeSelection(api_get_integer(ctx, args[1]));
	return 0;
}

// sizer

static long sizer_flags(void *ctx, void* arg)
{
	zsOptions ops[] = {
		{ "top", wxTOP },
		{ "bottom", wxBOTTOM },
		{ "left", wxLEFT },
		{ "right", wxRIGHT },
		{ "all", wxALL },
		{ "expand", wxEXPAND },
		{ "shaped", wxSHAPED },
		{ "minsize", wxFIXED_MINSIZE },
		{ "aligncenter", wxALIGN_CENTER },
		{ "alignvertical", wxALIGN_CENTER_VERTICAL },
		{ "alignhorizontal", wxALIGN_CENTER_HORIZONTAL },
		{ "alignleft", wxALIGN_LEFT },
		{ "alignright", wxALIGN_RIGHT },
		{ "aligntop", wxALIGN_TOP },
		{ "alignbottom", wxALIGN_BOTTOM },
		{ 0, 0 } };
	return get_options(ctx, api_get_string(ctx, arg), ops);
}

static void* sizer_add(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxSizer* sizer = (wxSizer*)api_get_user(ctx, args[0], SIZER_TYPE);
	if (api_is_integer(args[1])) {
		if (nargs < 3) api_input_error(ctx);
		int w = api_get_integer(ctx, args[1]);
		int h = api_get_integer(ctx, args[2]);
		sizer->Add(w, h, 0, 0);
		sizer->Layout();
		return 0;
	}
	int flag = 0;
	int prop = 0;
	if (nargs > 2) {
		flag = sizer_flags(ctx, args[2]);
		if (nargs > 3) prop = api_get_integer(ctx, args[3]);
	}
	if (api_get_type(args[1]) == SIZER_TYPE) {
		wxSizer *w = (wxSizer*)api_get_ptr(ctx, args[1]);
		sizer->Add(w, prop, flag);
		sizer->Layout();
	}
	else if (api_get_type(args[1]) == WINDOW_TYPE) {
		wxWindow *w = (wxWindow*)api_get_ptr(ctx, args[1]);
		sizer->Add(w, prop, flag);
		sizer->Layout();
	}
	else if (api_get_type(args[1]) == GRID_TYPE) {
		wxGrid *w = (wxGrid*)api_get_ptr(ctx, args[1]);
		sizer->Add(w, prop, flag);
		sizer->Layout();
	}
	else {
		api_input_error(ctx);
	}
	return 0;
}

static void* sizer_sizer(void *ctx, int nargs, void** args)
{
	if (nargs < 2) api_input_error(ctx);
	wxSizer *topsizer = (wxSizer*)api_get_user(ctx, args[0], SIZER_TYPE);
	const char *type = api_get_string(ctx, args[1]);
	int flag = 0;
	int prop = 0;
	if (strcmp(type, "horbox") == 0) {
		wxBoxSizer *sizer = new wxBoxSizer(wxHORIZONTAL);
		if (nargs > 2) {
			flag = sizer_flags(ctx, args[2]);
			if (nargs > 3) prop = api_get_integer(ctx, args[3]);
		}
		topsizer->Add(sizer, prop, flag);
		topsizer->Layout();
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	else if (strcmp(type, "verbox") == 0) {
		wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
		if (nargs > 2) {
			flag = sizer_flags(ctx, args[2]);
			if (nargs > 3) prop = api_get_integer(ctx, args[3]);
		}
		topsizer->Add(sizer, prop, flag);
		topsizer->Layout();
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	else if (strcmp(type, "flexgrid") == 0) {
		if (nargs < 6) api_input_error(ctx);
		int rows = api_get_integer(ctx, args[2]);
		int vgap = api_get_integer(ctx, args[3]);
		int cols = api_get_integer(ctx, args[4]);
		int hgap = api_get_integer(ctx, args[5]);
		wxFlexGridSizer *sizer = new wxFlexGridSizer(rows, cols, vgap, hgap);
		if (nargs > 6) {
			flag = sizer_flags(ctx, args[6]);
			if (nargs > 7) prop = api_get_integer(ctx, args[7]);
		}
		topsizer->Add(sizer, prop, flag);
		topsizer->Layout();
		return api_create_user(ctx, sizer, 0, 0, SIZER_TYPE);
	}
	api_runtime_error(ctx, "unsupported sizer type");
	return 0;
}

static void* sizer_update(void *ctx, int nargs, void** args)
{
	if (nargs < 1) api_input_error(ctx);
	wxSizer* sizer = (wxSizer*)api_get_user(ctx, args[0], SIZER_TYPE);
	sizer->Layout();
	return 0;
}

// the master event handler

static void* win_push(void *ctx, int nargs, void** args)
{
	if (nargs < 3) api_input_error(ctx);
	wxWindow *w = (wxWindow*)api_get_user(ctx, args[0], WINDOW_TYPE);
	void *func = api_get_func(ctx, api_get_string(ctx, args[1]));
	myEvtHandler* h = new myEvtHandler(ctx, func, w);
	myFrame *frame = (myFrame*)wxGetApp().GetTopWindow();
	frame->evth(h);
	w->PushEventHandler(h);
	const char* ev = api_get_string(ctx, args[2]);
	if (strcmp(ev, "key") == 0) {
		w->Connect(wxEVT_KEY_UP, (wxObjectEventFunction)(&myEvtHandler::OnKey), NULL, h);
		return 0;
	}
	if (strcmp(ev, "size") == 0) {
		w->Connect(wxEVT_SIZE, (wxObjectEventFunction)(&myEvtHandler::OnSize), NULL, h);
		return 0;
	}
	if (strcmp(ev, "paint") == 0) {
		w->Connect(wxEVT_PAINT, (wxObjectEventFunction)(&myEvtHandler::OnPaint), NULL, h);
		return 0;
	}
	if (strcmp(ev, "combo") == 0) {
		w->Connect(wxEVT_COMMAND_COMBOBOX_SELECTED , (wxObjectEventFunction)(&myEvtHandler::OnSelect), NULL, h);
		return 0;
	}
	if (strcmp(ev, "mouse") == 0) {
		w->Connect(wxEVT_LEFT_UP,       (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_LEFT_DOWN,     (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_LEFT_DCLICK,   (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_RIGHT_UP,      (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_RIGHT_DOWN,    (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_RIGHT_DCLICK,  (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_MIDDLE_UP,     (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_MIDDLE_DOWN,   (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_MIDDLE_DCLICK, (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_MOTION,        (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_ENTER_WINDOW,  (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_LEAVE_WINDOW,  (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		w->Connect(wxEVT_MOUSEWHEEL,    (wxObjectEventFunction)(&myEvtHandler::OnMouse), NULL, h);
		return 0;
	}
	if (nargs < 4) api_input_error(ctx);
	if (strcmp(ev, "menu") == 0) {
		for (int k = 3; k < nargs; k++) {
			if (api_is_integer(args[3])) {
				w->Connect(api_get_integer(ctx, args[3]), wxEVT_COMMAND_MENU_SELECTED, (wxObjectEventFunction)(&myEvtHandler::OnCommand), NULL, h);
			}
			else {
				int i, n = api_get_array_size(ctx, args[k]);
				for (i = 0; i < n; i++) {
					void *value = api_get_array_object2(ctx, args[k], i);
					w->Connect(api_get_integer(ctx, value), wxEVT_COMMAND_MENU_SELECTED, (wxObjectEventFunction)(&myEvtHandler::OnCommand), NULL, h);
				}
			}
		}
		return 0;
	}
	if (strcmp(ev, "tool") == 0) {
		if (api_is_integer(args[3])) {
			w->Connect(api_get_integer(ctx, args[3]), wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)(&myEvtHandler::OnCommand), NULL, h);
		}
		else {
			int i, n = api_get_array_size(ctx, args[3]);
			for (i = 0; i < n; i++) {
				void *value = api_get_array_object2(ctx, args[3], i);
				w->Connect(api_get_integer(ctx, value), wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)(&myEvtHandler::OnCommand), NULL, h);
			}
		}
		return 0;
	}
	return 0;
}

// event functions

void myEvtHandler::OnCommand(wxCommandEvent& event)
{
	if (_func) {
		void *p[2] = {
			api_create_integer(_ctx, event.GetId()),
			api_create_user(_ctx, _obj, 0, 0, WINDOW_TYPE)
		};
		api_call_func(_ctx, _func, 2, p);
	}
//	event.Skip();
}

void myEvtHandler::OnKey(wxKeyEvent& event)
{
	if (_func) {
		void *p[5] = {  api_create_integer(_ctx, event.GetKeyCode()),
		                api_create_integer(_ctx, event.ControlDown()),
			            api_create_integer(_ctx, event.AltDown()),
				        api_create_integer(_ctx, event.ShiftDown()),
						api_create_user(_ctx, _obj, 0, 0, WINDOW_TYPE)
		};
		api_call_func(_ctx, _func, 5, p);
	}
//	event.Skip();
}

void myEvtHandler::OnMouse(wxMouseEvent& event)
{
	if (_func) {
		void *p[7];
		WXTYPE type = event.GetEventType();
		if      (type == wxEVT_LEFT_DOWN)     p[0] = api_create_integer(_ctx, 1);
		else if (type == wxEVT_LEFT_UP)       p[0] = api_create_integer(_ctx, 2);
		else if (type == wxEVT_LEFT_DCLICK)   p[0] = api_create_integer(_ctx, 3);
		else if (type == wxEVT_MIDDLE_DOWN)   p[0] = api_create_integer(_ctx, 4);
		else if (type == wxEVT_MIDDLE_UP)     p[0] = api_create_integer(_ctx, 5);
		else if (type == wxEVT_MIDDLE_DCLICK) p[0] = api_create_integer(_ctx, 6);
		else if (type == wxEVT_RIGHT_DOWN)    p[0] = api_create_integer(_ctx, 7);
		else if (type == wxEVT_RIGHT_UP)      p[0] = api_create_integer(_ctx, 8);
		else if (type == wxEVT_RIGHT_DCLICK)  p[0] = api_create_integer(_ctx, 9);
		else if (type == wxEVT_MOTION)        p[0] = api_create_integer(_ctx, 10);
		else                                  p[0] = api_create_integer(_ctx, 0);
		p[1] = api_create_integer(_ctx, event.GetX());
		p[2] = api_create_integer(_ctx, event.GetY());
		p[3] = api_create_integer(_ctx, event.ControlDown());
		p[4] = api_create_integer(_ctx, event.AltDown());
		p[5] = api_create_integer(_ctx, event.ShiftDown());
		p[6] = api_create_user(_ctx, _obj, 0, 0, WINDOW_TYPE);
		api_call_func(_ctx, _func, 7, p);
	}
//	event.Skip();
}

void myEvtHandler::OnSize(wxSizeEvent& event)
{
	if (_func) {
		wxWindow *w = (wxWindow*)event.GetEventObject();
		wxRect rect = w->GetClientRect();
		void *p[2] = {  api_create_integer(_ctx, rect.GetWidth()),
						api_create_integer(_ctx, rect.GetHeight()) };
		api_call_func(_ctx, _func, 2, p);
	}
	event.Skip();
}

void myEvtHandler::OnPaint(wxPaintEvent& event)
{
	if (_func) {
		wxWindow *w = (wxWindow*)event.GetEventObject();
		void *p[1] = {  api_create_user(_ctx, w, 0, 0, WINDOW_TYPE) };
		api_call_func(_ctx, _func, 1, p);
	}
	event.Skip();
}

void myEvtHandler::OnSelect(wxCommandEvent& event)
{
	if (_func) {
		void *p[1] = { api_create_user(_ctx, _obj, 0, 0, WINDOW_TYPE) };
		api_call_func(_ctx, _func, 1, p);
	}
//	event.Skip();
}

void myTreeCtrl::OnActivated(wxTreeEvent& event)
{
	if (_func) {
		wxTreeItemId id = event.GetItem();
		wxString s = GetItemText(id);
		id = GetItemParent(id);
		while (id.IsOk()) {
			s = GetItemText(id) + "/" + s;
			id = GetItemParent(id);
		}
		std::string str = wxs2str(s);
		void *p[2] = { api_create_string(_ctx, "ACTIVATE"),
					   api_create_string(_ctx, str.c_str()) };
		api_call_func(_ctx, _func, 2, p);
	}
//	event.Skip();
}

void myTreeCtrl::OnRightClick(wxTreeEvent& event)
{
	if (_func) {
		wxTreeItemId id = event.GetItem();
		wxString s = GetItemText(id);
		id = GetItemParent(id);
		while (id.IsOk()) {
			s = GetItemText(id) + "/" + s;
			id = GetItemParent(id);
		}
		std::string str = wxs2str(s);
		void *p[2] = { api_create_string(_ctx, "RIGHTCLICK"),
					   api_create_string(_ctx, s.c_str()) };
		api_call_func(_ctx, _func, 2, p);
	}
//	event.Skip();
}

void myTreeCtrl::OnBeginDrag(wxTreeEvent& event)
{
	if (_func) {
		wxTreeItemId id = event.GetItem();
		wxString s = GetItemText(id);
		id = GetItemParent(id);
		while (id.IsOk()) {
			s = GetItemText(id) + "/" + s;
			id = GetItemParent(id);
		}
		std::string str = wxs2str(s);
		void *p[2] = { api_create_string(_ctx, "BEGINDRAG"),
					   api_create_string(_ctx, s.c_str()) };
		api_call_func(_ctx, _func, 2, p);
	}
//	event.Skip();
}

void myTreeCtrl::OnEndDrag(wxTreeEvent& event)
{
	if (_func) {
		wxTreeItemId id = event.GetItem();
		wxString s = GetItemText(id);
		id = GetItemParent(id);
		while (id.IsOk()) {
			s = GetItemText(id) + "/" + s;
			id = GetItemParent(id);
		}
		std::string str = wxs2str(s);
		void *p[2] = { api_create_string(_ctx, "ENDDRAG"),
					   api_create_string(_ctx, s.c_str()) };
		api_call_func(_ctx, _func, 2, p);
	}
//	event.Skip();
}

void myFrame::OnSize(wxSizeEvent& event)
{
	if (_func) {
		int w, h;
		GetSize(&w, &h);
		void *p[3] = {  api_create_string(_ctx, "size"),
						api_create_integer(_ctx, w),
						api_create_integer(_ctx, h) };
		api_call_func(_ctx, _func, 3, p);
	}
	event.Skip();
}


void myFrame::OnTimer(wxTimerEvent& event)
{
	if (_func) {
		void *p[2] = {  api_create_string(_ctx, "timer"),
						api_create_integer(_ctx, timer().GetInterval())
						};
		api_call_func(_ctx, _func, 2, p);
	}
//	event.Skip();
}


void myFrame::OnDropFile(wxDropFilesEvent& event)
{
	if (_func) {
		std::string str = wxs2str(*event.GetFiles());
		void *p[3] = {  api_create_string(_ctx, "dropfile"),
                        api_create_string(_ctx, str.c_str()),
                        api_create_integer(_ctx, event.GetNumberOfFiles()) };
		api_call_func(_ctx, _func, 3, p);
	}
	event.Skip();
}


void myFrame::OnPaint(wxPaintEvent& event)
{
	wxToolBar *toolbar = GetToolBar();
	if (toolbar) toolbar->Realize();

	if (_func) {
		wxPaintDC dc(this);
		void *p[2] = {  api_create_string(_ctx, "paint"),
                        api_create_user(_ctx, GetHandle(), 0, 0, 0) };
		api_call_func(_ctx, _func, 2, p);
	}
	event.Skip();
}

void myFrame::OnActivate(wxActivateEvent& event)
{
	if (_func) {
		void *p[3] = {  api_create_string(_ctx, "activate"),
			api_create_user(_ctx, GetHandle(), 0, 0, 0),
			api_create_integer(_ctx, event.GetActive()?1:0) };
		api_call_func(_ctx, _func, 2, p);
	}
	event.Skip();
}

void myFrame::OnErase(wxEraseEvent& event)
{
	if (_func) {
		void *p[2] = {  api_create_string(_ctx, "erase"),
                        api_create_user(_ctx, GetHandle(), 0, 0, 0) };
		api_call_func(_ctx, _func, 2, p);
	}
	event.Skip();
}

void myFrame::OnFocus(wxFocusEvent& event)
{
	if (_func) {
		void *p[2] = {  api_create_string(_ctx, "focus"),
                        api_create_user(_ctx, GetHandle(), 0, 0, 0) };
		api_call_func(_ctx, _func, 2, p);
	}
	event.Skip();
}

/////////////////////////////////////////////////////////////////////


class zswRegPrimitive
{
public:
	zswRegPrimitive()
	{
		api_add_primitive("wxmsg",		0,				wx_message);
		api_add_primitive("wxframe",	0,				wx_frame);
		api_add_primitive("wxcast",		0,				wx_cast);
		api_add_primitive("wxactive",	0,				wx_active);
		api_add_primitive("wxsize",		0,				wx_size);
		api_add_primitive("wxexec",		0,				wx_exec);
		api_add_primitive("wxexit",		0,				wx_exit);
		api_add_primitive("wxsplit",	0,				wx_split);

		api_add_primitive("timer",		TOPWIN_TYPE,	topwin_timer);
		api_add_primitive("directory",	TOPWIN_TYPE,	topwin_dirdialog);
		api_add_primitive("file",		TOPWIN_TYPE,	topwin_filedialog);
		api_add_primitive("color",		TOPWIN_TYPE,	topwin_colordialog);
		api_add_primitive("font",		TOPWIN_TYPE,	topwin_fontdialog);
		api_add_primitive("icon",		TOPWIN_TYPE,	topwin_icon);
		api_add_primitive("title",		TOPWIN_TYPE,	topwin_title);;
		api_add_primitive("size",		TOPWIN_TYPE,	topwin_size);
		api_add_primitive("menubar",	TOPWIN_TYPE,	topwin_menubar);
		api_add_primitive("toolbar",	TOPWIN_TYPE,	topwin_toolbar);
		api_add_primitive("statusbar",	TOPWIN_TYPE,	topwin_statusbar);
		api_add_primitive("push",		TOPWIN_TYPE,	topwin_push);
		api_add_primitive("pop",		TOPWIN_TYPE,	topwin_pop);
		api_add_primitive("update",		TOPWIN_TYPE,	topwin_update);
		api_add_primitive("evth",		TOPWIN_TYPE,	topwin_callback);

		api_add_primitive("enable",		WINDOW_TYPE,	win_enable);
		api_add_primitive("disable",	WINDOW_TYPE,	win_disable);
		api_add_primitive("button",		WINDOW_TYPE,	win_button);
		api_add_primitive("bitmap",		WINDOW_TYPE,	win_bitmap);
		api_add_primitive("bgcolor",	WINDOW_TYPE,	win_bgcolor);
		api_add_primitive("checkbox",	WINDOW_TYPE,	win_checkbox);
		api_add_primitive("close",		WINDOW_TYPE,	win_close);
		api_add_primitive("combobox",	WINDOW_TYPE,	win_combobox);
		api_add_primitive("cursor",		WINDOW_TYPE,	win_cursor);
		api_add_primitive("fgcolor",	WINDOW_TYPE,	win_fgcolor);
		api_add_primitive("font",		WINDOW_TYPE,	win_font);
		api_add_primitive("focus",		WINDOW_TYPE,	win_focus);
		api_add_primitive("get",		WINDOW_TYPE,	win_get);
		api_add_primitive("grid",		WINDOW_TYPE,	win_grid);
		api_add_primitive("hwnd",		WINDOW_TYPE,	win_hwnd);
		api_add_primitive("id",			WINDOW_TYPE,	win_id);
		api_add_primitive("notebook",	WINDOW_TYPE,	win_notebook);
		api_add_primitive("panel",		WINDOW_TYPE,	win_panel);
		api_add_primitive("push",		WINDOW_TYPE,	win_push);
		api_add_primitive("radiobox",	WINDOW_TYPE,	win_radiobox);
		api_add_primitive("html",		WINDOW_TYPE,	win_html);
		api_add_primitive("set",		WINDOW_TYPE,	win_set);
		api_add_primitive("size",		WINDOW_TYPE,	win_size);
		api_add_primitive("sizer",		WINDOW_TYPE,	win_sizer);
		api_add_primitive("slider",		WINDOW_TYPE,	win_slider);
		api_add_primitive("splitter",	WINDOW_TYPE,	win_splitter);
		api_add_primitive("statictext",	WINDOW_TYPE,	win_statictext);
		api_add_primitive("style",		WINDOW_TYPE,	win_style);
		api_add_primitive("textbox",	WINDOW_TYPE,	win_textbox);
		api_add_primitive("treeview",	WINDOW_TYPE,	win_treeview);
		api_add_primitive("refresh",	WINDOW_TYPE,	win_refresh);
		api_add_primitive("position",	WINDOW_TYPE,	win_position);
		api_add_primitive("wrap",		WINDOW_TYPE,	win_wrap);

		api_add_primitive("load",		HTML_TYPE,		html_load);
		api_add_primitive("font",		HTML_TYPE,		html_font);

		api_add_primitive("min",		SPLITTER_TYPE,	splitter_min);
		api_add_primitive("horizontal",	SPLITTER_TYPE,	splitter_horizontal);
		api_add_primitive("vertical",	SPLITTER_TYPE,	splitter_vertical);

		api_add_primitive("__get",		GRID_TYPE,		grid_get);
		api_add_primitive("__set",		GRID_TYPE,		grid_set);
		api_add_primitive("label",		GRID_TYPE,		grid_label);
		api_add_primitive("rows",		GRID_TYPE,		grid_rows);
		api_add_primitive("cols",		GRID_TYPE,		grid_cols);
		api_add_primitive("clear",		GRID_TYPE,		grid_clear);
		api_add_primitive("width",		GRID_TYPE,		grid_width);
		api_add_primitive("height",		GRID_TYPE,		grid_height);
		api_add_primitive("enable",		GRID_TYPE,		grid_enable);
		api_add_primitive("disable",	GRID_TYPE,		grid_disable);
		api_add_primitive("readonly",	GRID_TYPE,		grid_readonly);

		api_add_primitive("root",		TREE_TYPE,		tree_root);
		api_add_primitive("append",		TREE_TYPE,		tree_append);

		api_add_primitive("add",		NOTEBOOK_TYPE,	notebook_add);
		api_add_primitive("select",		NOTEBOOK_TYPE,	notebook_select);

		api_add_primitive("add",		SIZER_TYPE,		sizer_add);
		api_add_primitive("sizer",		SIZER_TYPE,		sizer_sizer);
		api_add_primitive("update",		SIZER_TYPE,		sizer_update);

		api_add_primitive("check",		MENUBAR_TYPE,	menubar_check);
		api_add_primitive("enable",		MENUBAR_TYPE,	menubar_enable);
		api_add_primitive("label",		MENUBAR_TYPE,	menubar_label);
	}
};

static zswRegPrimitive wx_register;